from lib389._constants import *
from lib389.properties import *
from lib389.tasks import *
from lib389.utils import *
from lib389.schema import Schema
from flask import request, jsonify
from flask import Blueprint

from .Authentication import authenticateConnection
from .Helpers import jsonResponse

schema_api = Blueprint('schema_api', __name__)


"""
Schema Representations::

    Attributes:
    {
        'names': [name, name,...],
        'desc': "val",
        'oid': "val",
        'ordering': "val",
        'single_value': "val",
        'substr': "val",
        'equality: "val",
        'syntax': "syntax oid",
        'usage': "0|1|2|3", [userApplications, directoryOperation,
                             distributedOperation, dSAOperation]
        'obsolete': "0|1",
        'no_user_mod': attribute.no_user_mod,
        'sup': ["superior", "superior"]
    }

    Objectclasses:
    {
        'names': [name, name,...],
        'desc': "val",
        'oid': "val",
        'obsolete': "0|1",
        'kind': "0|1|2",  [structure, abstract, auxilary]
        'may': ["attr", "attr"]
        'must': ["attr", "attr"]
    }

    MatchingRules:
    {
        'oid': OID,
        'names': [name, name]
        'desc': description
        'obsolete': "0|1"
        'syntax': "syntax oid"
    }
"""

USAGE = ['userApplications', 'directoryOperation', 'distributedOperation',
         'dSAOperation']
KIND = ['STRUCTURAL', 'ABSTRACT', 'AUXILIARY']


def jsonAttrToStr(rep):
    """
    Take our JSON Attribute rep and turn it into the string value
    """
    attrVal = '('
    if 'oid' in rep:
        attrVal += '%s ' % str(rep['oid'])
    if 'names' in rep:
        if len(rep['names']) > 1:
            attrVal += 'NAME ('
            for name in rep['names']:
                attrVal += (" '%s'" % name)
            attrVal += (' ) ')
        else:
            attrVal += ("NAME '%s' " % rep['names'][0])
    if 'desc' in rep:
        attrVal += ("DESC '%s' " % rep['desc'])
    if 'ordering' in rep:
        attrVal += "ORDERING %s " % rep['ordering']
    if 'sup' in rep:
        if len(rep['sup']) > 1:
            attrVal += ('SUP (')
            for name in rep['sup']:
                attrVal += (" '%s'" % name)
            attrVal += (' ) ')
        else:
            attrVal += ("SUP '%s' " % rep['sup'][0])
    if 'equality' in rep:
        attrVal += ("EQUALITY %s " % rep['equality'])
    if 'substr' in rep:
        attrVal += ("SUBSTR %s " % rep['substr'])
    if 'syntax' in rep:
        attrVal += ("SYNTAX %s " % rep['syntax'])
    if 'single_value' in rep:
        attrVal += ("SINGLE-VALUE ")
    if 'obsolete' in rep and int(rep['obsolete']) > 0:
        attrVal += ("OBSOLETE ")
    if 'usage' in rep:
        attrVal += ("USAGE %s " % USAGE[int(str(rep['usage']))])
    if 'no_user_mod' in rep and int(rep['no_user_mod']) > 0:
        attrVal += ("NO-USER-MODIFICATION ")
    attrVal += (')')

    return str(attrVal)


def jsonOCToStr(rep):
    """
    Take our JSON ObjectClass rep and turn it into the string value
    """
    ocVal = '('
    if 'oid' in rep:
        ocVal += (rep['oid'])
    if 'names' in rep:
        if len(rep['names']) > 1:
            ocVal += ('NAME (')
            for name in rep['names']:
                ocVal += (" '%s'" % name)
            ocVal += (' ) ')
        else:
            ocVal += ("NAME '%s' " % rep['names'][0])
    if 'desc' in rep:
        ocVal += ("DESC '%s' " % rep['desc'])
    if 'sup' in rep:
        if len(rep['sup']) > 1:
            ocVal += ('SUP (')
            for name in rep['sup']:
                ocVal += (" '%s'" % name)
            ocVal += (' ) ')
        else:
            ocVal += ("SUP '%s' " % rep['sup'][0])

    if 'kind' in rep:
        ocVal += ("%s " % KIND[int(str(rep['kind']))])
    if 'obsolete' in rep and int(rep['obsolete']) > 0:
        ocVal += ("OBSOLETE ")
    if 'may' in rep:
        ocVal += ('MAY (')
        firstAttr = True
        for attr in rep['may']:
            if firstAttr:
                ocVal += (" %s" % attr)
                firstAttr = False
            else:
                ocVal += (" $ %s" % attr)
        ocVal += (' ) ')

    if 'must' in rep:
        ocVal += ('MUST (')
        firstAttr = True
        for attr in rep['must']:
            if firstAttr:
                ocVal += (" %s" % attr)
                firstAttr = False
            else:
                ocVal += (" $ %s" % attr)
        ocVal += (' ) ')

    ocVal += (')')

    return str(ocVal)


@schema_api.route('/v1/schema/attributes',
                  methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def schema_attributes_list(ds):
    """
    Get all the attributes
    """
    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {'allow': 'GET OPTIONS HEAD'}
        return jsonify(msg)

    try:
        # Should get the results before for error handling.
        attributes = ds.schema.get_attributetypes()
        results = []
        for attribute in attributes:
            results.append({
                'names': attribute.names,
                'desc': attribute.desc,
                'oid': attribute.oid,
                'ordering': attribute.ordering,
                'single_value': attribute.single_value,
                'substr': attribute.substr,
                'equality': attribute.equality,
                'syntax': attribute.syntax,
                'usage': attribute.usage,
                'obsolete': attribute.obsolete,
                'no_user_mod': attribute.no_user_mod,
                'sup': attribute.sup
            })
    except ldap.LDAPError as e:
        return jsonResponse("Error getting attributes: " % str(e), 403,
                            request.url)

    if request.method == 'HEAD':
        #
        # Return the HEAD Response
        #
        if attributes is None:
            return jsonResponse("", 403, request.url)
        return jsonResponse("", 200, request.url)

    if request.method == 'GET':
        if attributes is None:
            return jsonResponse("Failed to search for attribute: %s" % str(e),
                                403, request.url)
        return jsonResponse(results, 200, request.url)


@schema_api.route('/v1/schema/attributes/<attribute_name>',
                  methods=['GET', 'PUT', 'DELETE', 'OPTIONS', 'HEAD'])
@authenticateConnection
def schema_attribute(ds, attribute_name):
    """
    Get a particular attribute
    """
    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {'allow': 'GET PUT DELETE OPTIONS HEAD'}
        return jsonify(msg)

    try:
        result = ds.schema.query_attributetype(attribute_name)
    except ldap.LDAPError as e:
        return jsonResponse("Error getting attribute: " % str(e), 403,
                            request.url)

    if request.method == 'HEAD':
        #
        # Return the HEAD Response
        #
        if result is None:
            return jsonResponse("", 404, request.url)
        return jsonResponse("", 200, request.url)

    if request.method == 'GET':
        if result is None:
            return jsonResponse("Failed to find attribute", 404, request.url)
        # Split the tuple
        attribute, must, may = result
        results = {'names': attribute.names,
                   'desc': attribute.desc,
                   'oid': attribute.oid,
                   'ordering': attribute.ordering,
                   'single_value': attribute.single_value,
                   'equality': attribute.equality,
                   'substr': attribute.substr,
                   'syntax': attribute.syntax,
                   'usage': attribute.usage,
                   'obsolete': attribute.obsolete,
                   'no_user_mod': attribute.no_user_mod,
                   'sup': attribute.sup,
                   'must': map(lambda x: x.names[0], must),
                   'may': map(lambda x: x.names[0], may)}
        return jsonResponse(results, 200, request.url)

    if request.method == 'PUT':
        json_req = request.get_json(force=True)
        if not json_req:
            return jsonResponse("JSON representation missing", 403,
                                request.url)
        newattr = jsonAttrToStr(json_req)

        attrObj = Schema(ds)
        try:
            attrObj.add_attribute(newattr)
        except ldap.TYPE_OR_VALUE_EXISTS:
            # idempotent - return the entry as is
            return jsonResponse(json_req, 201,
                                request.url + '/' + json_req['names'][0])

        except ldap.LDAPError as e:
            return jsonResponse('Failed to add attribute: %s' % str(e), 403,
                                request.url)

        return jsonResponse(json_req, 201,
                            request.url + '/' + json_req['names'][0])

    if request.method == 'DELETE':
        schemaObj = Schema(ds)
        try:
            delete_val = None
            name_cmp = "'%s'" % attribute_name

            entry = schemaObj.get_entry()
            attributes = entry.getValues('attributeTypes')
            for val in attributes:
                if name_cmp.lower() in val.lower():
                    delete_val = val
                    break

            if delete_val:
                ds.modify_s(DN_SCHEMA, [(ldap.MOD_DELETE, 'attributeTypes',
                            delete_val)])
                return jsonResponse('', 200, request.url)
            else:
                # Idempotent - no val, already deleted, return success
                return jsonResponse('', 200, request.url)

        except ldap.LDAPError as e:
            return jsonResponse('Failed to delete attribute: %s' % str(e), 403,
                                request.url)


@schema_api.route('/v1/schema/objectclasses',
                  methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def schema_objectclasses_list(ds):
    """
    Get all the objectclasses
    """
    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {'allow': 'GET OPTIONS HEAD'}
        return jsonify(msg)

    try:
        objectclasses = ds.schema.get_objectclasses()
    except ldap.LDAPError as e:
        return jsonResponse("Error getting objectclasses: " % str(e), 403,
                            request.url)

    if request.method == 'HEAD':
        #
        # Return the HEAD Response
        #
        if objectclasses:
            return jsonResponse("", 200, request.url)
        else:
            return jsonResponse("", 404, request.url)

    if request.method == 'GET':
        results = []
        for objectclass in objectclasses:

            results.append({
                'names': objectclass.names,
                'desc': objectclass.desc,
                'kind': objectclass.kind,
                'may': objectclass.may,
                'must': objectclass.must,
                'obsolete': objectclass.obsolete,
                'oid': objectclass.oid,
                'schema_attribute': objectclass.schema_attribute,
                'sup': objectclass.sup,
            })

        return jsonResponse(results, 200, request.url)


@schema_api.route('/v1/schema/objectclasses/<objectclass_name>',
                  methods=['GET', 'PUT', 'DELETE', 'OPTIONS', 'HEAD'])
@authenticateConnection
def schema_objectclass(ds, objectclass_name):
    """
    Get a particular objectclass
    """
    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {'allow': 'GET PUT DELETE OPTIONS HEAD'}
        return jsonify(msg)

    try:
        objectclass = ds.schema.query_objectclass(objectclass_name)
    except ldap.LDAPError as e:
        return jsonResponse("Error getting objectclass: " % str(e), 403,
                            request.url)

    # This does the right thing regardless of resource type

    if request.method == 'HEAD':
        #
        # Return the HEAD Response
        #
        if objectclass is None:
            return jsonResponse("", 404, request.url)
        return jsonResponse("", 200, request.url)

    if request.method == 'GET':
        if objectclass is None:
            return jsonResponse("Failed to find objectclass", 404, request.url)
        results = {
            'names': objectclass.names,
            'desc': objectclass.desc,
            'kind': objectclass.kind,
            'may': objectclass.may,
            'must': objectclass.must,
            'obsolete': objectclass.obsolete,
            'oid': objectclass.oid,
            'schema_attribute': objectclass.schema_attribute,
            'sup': objectclass.sup,
        }
        return jsonResponse(results, 200, request.url)

    if request.method == 'PUT':
        json_req = request.get_json(force=True)
        if not json_req:
            return jsonResponse("JSON representation missing", 403,
                                request.url)
        newoc = jsonOCToStr(json_req)

        attrObj = Schema(ds)
        try:
            attrObj.add_objectclass(newoc)
        except ldap.TYPE_OR_VALUE_EXISTS:
            # idempotent - return the entry as is
            return jsonResponse(json_req, 201,
                                request.url + '/' + json_req['names'][0])
        except ldap.LDAPError as e:
            return jsonResponse('Failed to add objectclass: %s' % str(e), 403,
                                request.url)

        return jsonResponse(json_req, 201,
                            request.url + '/' + json_req['names'][0])

    if request.method == 'DELETE':
        schemaObj = Schema(ds)
        try:
            delete_val = None
            name_cmp = "'%s'" % objectclass_name

            entry = schemaObj.get_entry()
            ocs = entry.getValues('objectClasses')
            for val in ocs:
                if name_cmp.lower() in val.lower():
                    delete_val = val
                    break

            if delete_val:
                ds.modify_s(DN_SCHEMA, [(ldap.MOD_DELETE, 'objectClasses',
                            delete_val)])
                return jsonResponse('', 200, request.url)
            else:
                # idempotent, return success
                return jsonResponse('', 200, request.url)

        except ldap.LDAPError as e:
            return jsonResponse('Failed to delete objectclass: %s' % str(e),
                                403, request.url)


@schema_api.route('/v1/schema/matchingrules',
                  methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def schema_matchingrules(ds):
    """
    Get all the matching rules
    """
    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {'allow': 'GET OPTIONS HEAD'}
        return jsonify(msg)

    try:
        matchingRules = ds.schema.get_matchingrules()
    except ldap.LDAPError as e:
        return jsonResponse("Error getting matching rules: " % str(e), 403,
                            request.url)

    if request.method == 'HEAD':
        #
        # Return the HEAD Response
        #
        if matchingRules is None:
            return jsonResponse("", 404, request.url)
        return jsonResponse("", 200, request.url)

    if request.method == 'GET':
        results = []
        for mr in matchingRules:
            results.append({
                'names': mr.names,
                'oid': mr.oid,
                'desc': mr.desc,
                'obsolete': mr.obsolete,
                'syntax': mr.syntax
            })

        return jsonResponse(results, 200, request.url)


@schema_api.route('/v1/schema/matchingrules/<mr_name>',
                  methods=['GET', 'OPTIONS', 'HEAD'])
@authenticateConnection
def schema_matchingrule(ds, mr_name):
    """
    Get a particular matching rule
    """
    if request.method == 'OPTIONS':
        #
        # Return a list of the allowed methods
        #
        msg = {'allow': 'GET OPTIONS HEAD'}
        return jsonify(msg)

    try:
        matchingRule = ds.schema.query_matchingrule(mr_name)
    except ldap.LDAPError as e:
        return jsonResponse("Error getting matching rule: " % str(e), 403,
                            request.url)

    if request.method == 'HEAD':
        #
        # Return the HEAD Response
        #
        if matchingRule is None:
            return jsonResponse("", 404, request.url)
        return jsonResponse("", 200, request.url)

    if request.method == 'GET':
        if matchingRule is None:
            return jsonResponse("Failed to find matching rule", 404,
                                request.url)
        results = {
            'names': matchingRule.names,
            'desc': matchingRule.desc,
            'obsolete': matchingRule.obsolete,
            'oid': matchingRule.oid,
            'syntax': matchingRule.syntax,
        }
        return jsonResponse(results, 200, request.url)
